var mongoose = require('mongoose');
var Schema = mongoose.Schema;
mongoose.Promise = global.Promise;

/* An fs object
 */
var FileSystemSchema = new Schema({
	// _id     : Schema.Types.ObjectId,
	name    : String,
	created : {type: Date, default: Date.now },
	updated : Date,
	type    : { type: String, enum: ['directory', 'file'] },
	content : String,
	parent  : { type: Schema.Types.ObjectId, ref: 'FSObject' },
	nodes   : [{ type: String }] 
})

/* guarantee uniqueness of name - parent */
FileSystemSchema.index({ name: 1, parent: 1 }, { unique: true });


/* Retrieve all contents from the f/s
 */
FileSystemSchema.methods.getAllObjects = function(){
	return this.model('FSObject')
		.find({})
		.lean()
		.exec();
}

FileSystemSchema.methods.getFileObjectById = function(id){
	return this.model('FSObject')
		.findOne({ _id: id })
		.lean()
		.exec();
}

/* Overwrite the contents of a file
 * precondition: file object must be of type 'file'
 */
FileSystemSchema.methods.overwriteFile = function(id, newContent){
	return this.model('FSObject')
		.findOne({ _id : id })
		.update({ $set: { content: newContent, updated: new Date() }})
		.exec();
}

/* Return all the file objects with a matching name
 * @param name : name to match
 */
FileSystemSchema.methods.getObjectsByName = function(name){
	return this.model('FSObject')
		.find({ name : name })
		.lean()
		.exec();
}

/* Retrieve a file object based on its name and parentId
 * @param name     : name of the object
 * @param parentId : _id field of the parent for this object 
 * 
 */
FileSystemSchema.methods.getFileObject = function(name, parentId){
	console.log('<T> FSObject.js -- getFileObject - name: ' + name);
	console.log('<T> FSObject.js -- getFileObject - parentID: ' + parentId);
	return this.model('FSObject')
		.findOne({ name : name, parent: parentId })
		.lean()
		.exec();
}

/* Delete a file object and all its children
 */
FileSystemSchema.methods.deleteFileObject = function(id){
	return this.model('FSObject')
		.find({ $or: [{ _id: id }, { parent: id }] })
		.remove()
		.exec();
}

/* Retrieve the children of a directory 
 * @param id : _id field of the directory
 */
FileSystemSchema.methods.getChildren = function(){
	return this.model('FSObject')
		.find({ parent: this._id })
		.lean()
		.exec();
}

FileSystemSchema.statics.getChildrenOf = function(id){
	return this.model('FSObject')
		.find({ parent: id })
		.lean()
		.exec();
}

/* moves a file object
 * @param id          : _id field of the object
 * @param newParentId : new parent of the object 
 */
FileSystemSchema.methods.updateParent = function(id, newParentId){
	return this.model('FSObject')
		.find({ _id: id })
		.update({ $set: { parent: newParentId } })
		.exec();
}

/* rename a file object
 * @param id      : _id field of the object
 * @param newName : the new name of the object
 */
FileSystemSchema.methods.updateName = function(id, newName){
	return this.model('FSObject')
		.find({ _id: id })
		.update({ $set: { name: newName } })
		.exec();
}

FileSystemSchema.methods.getOffspring = function(rel_path){
	var self = this;
	console.log(rel_path);
	var tokens = rel_path.split('/');
	if (tokens[tokens.length-1] === '') tokens.pop();
	
	if (tokens.length === 0){
		return Promise.resolve(this);
	}
	else {
		return new Promise(function(resolve, reject){
			// console.log("Getting "+tokens[0]);
			self.model('FSObject')
				.findOne({ parent: self._id, name: tokens[0] })
				.exec(function(err, obj){
					if (obj){
						obj.getOffspring(tokens.slice(1).join('/'))
						.then(resolve, reject);	
					}
					else {
						reject("Not Found")
					}
				})
		}) 
	}
}

function getNode(abs_path){
	var tokens = abs_path.split('/');
	if (tokens[tokens.length-1] === '') tokens.pop();
	if (tokens.length === 1){
		return FSObject.getChildrenOf(null)
			.then(function(children){
				var lean = {
					_id: null,
					abs_path: abs_path,
					children: {}
				};
				children.forEach(function(child){
					lean.children[child.name] = child;
				})
				return lean;
			})
	}
	else {
		return new Promise(function(resolve, reject){
			FSObject.findOne({
				parent: null,
				name: tokens[1]
			}).exec(function(err, obj){
				if (obj){
					obj.getOffspring(tokens.slice(2).join('/'))
						.then(function(node){
							var lean = node.toObject();
							if (node.type === 'directory'){
								lean.children = {};
								node.getChildren()
								.then(function(children){
									children.forEach(function(child){
										lean.children[child.name] = child;
									})
									resolve(lean);
								})	
							}
							else {
								resolve(lean);
							}
						}, reject);
				}
				else {
					var error = new Error("Object Not Found");
					error.code = 'ENOENT';
					error.context = {
						filename: tokens[1],
						tokens: tokens
					}
					reject(error);
				}
			})
		})
	}
}
FileSystemSchema.statics.getNode = getNode;

var FSObject = mongoose.model('FSObject', FileSystemSchema);

module.exports = FSObject;